# This file was automatically generated by SWIG (http://www.swig.org).
# Version 3.0.5
#
# Do not make changes to this file unless you know what you are doing--modify
# the SWIG interface file instead.

package BitPay::KeyUtils;
use base qw(Exporter);
use base qw(DynaLoader);
package BitPay::KeyUtilsc;
bootstrap BitPay::KeyUtils;
package BitPay::KeyUtils;
@EXPORT = qw();

# ---------- BASE METHODS -------------

package BitPay::KeyUtils;

sub TIEHASH {
    my ($classname,$obj) = @_;
    return bless $obj, $classname;
}

sub CLEAR { }

sub FIRSTKEY { }

sub NEXTKEY { }

sub FETCH {
    my ($self,$field) = @_;
    my $member_func = "swig_${field}_get";
    $self->$member_func();
}

sub STORE {
    my ($self,$field,$newval) = @_;
    my $member_func = "swig_${field}_set";
    $self->$member_func($newval);
}

sub this {
    my $ptr = shift;
    return tied(%$ptr);
}


# ------- FUNCTION WRAPPERS --------

package BitPay::KeyUtils;

*generatePem = *BitPay::KeyUtilsc::generatePem;
*generateSinFromPem = *BitPay::KeyUtilsc::generateSinFromPem;
*getPublicKeyFromPem = *BitPay::KeyUtilsc::getPublicKeyFromPem;
*signMessageWithPem = *BitPay::KeyUtilsc::signMessageWithPem;
*bpSignMessageWithPem = *BitPay::KeyUtilsc::bpSignMessageWithPem;
*bpGeneratePem = *BitPay::KeyUtilsc::bpGeneratePem;
*bpGenerateSinFromPem = *BitPay::KeyUtilsc::bpGenerateSinFromPem;
*bpGetPublicKeyFromPem = *BitPay::KeyUtilsc::bpGetPublicKeyFromPem;

# ------- VARIABLE STUBS --------

package BitPay::KeyUtils;

1;
